/* Copyright (c) Colorado School of Mines, 2011.*/
/* All rights reserved.                       */

/* SMO0THINT2: $Revision: 1.11 $ ; $Date: 2012/04/19 21:09:31 $	*/

#include "par.h"

/*********************** self documentation ******************************/
char *sdoc[] = {
" 									",
" SMOOTHINT2 --- SMOOTH non-uniformly sampled INTerfaces, via the damped",
"  		least-squares technique					",
" 									",
"  smoothint2 <input ninf= >output [optional parameters]		",
" 									",
" Required Parameters:							",
" <input                 file containing original interfaces		",
" >output                file containing smoothed interfaces	 	",
" 									",
" Optional Parameters:							",
" ninf=5                number of interfaces  				",
" r=100			smoothing parameter 				",
" npmax=101		maximum number of points in interfaces		",
" 									",
" Notes:								",
" The input file is an ASCII file. Each interface is represented by pairs",
" (non-uniform sampling) of x and z values, with one pair of values on	",
" each line, separated by spaces or tabs. Each interface is separated with",
" an entry with a large negative z value for example: 1.0     -9999.	",
" There is no entry for the surface. The surface is assumed to be flat  ",
" with z=0.								",
" This is similar to a CSHOT model file without a surface entry and	",
" without comments.							",
" 									",
" The smoothing method is analogous to a moving window averaging process",
" (but not the same) with the parameter \"r\" being analogous to the \"width\"",
" of the window. Thus, the size of \"r\" must be chosen to by compatible",
" with the scale (wavelengths) of the variations of the interfaces in the",
" model being smoothed.							",
" 									",
" Example using the test data set generated by unif2: 			",
" unif2 tfile=tfilename							",
" Compare the unsmoothed interface model:				",
" unif2 < tfilename method=interpolation_method |			",
"	 			psimage n1=100 n2=100 d1=10 d2=10 | ...	",
" To the smoothed interface model:					",
" smoothint2 r=100 < tfilename | unif2 method=interpolation_method |	", 
"	psimage n1=100 n2=100 d1=10 d2=10 | ...				",
"   									",
NULL};

/*
 * Credits:
 *  CWP: Zhenyue Liu, Jan 1994
 * Reference:
 *    Liu, Zhenyue, 1994, Velocity smoothing: theory and implementation, 
 *    Project Review, 1994, Consortium Project on Seismic Inverse Methods
 *    for Complex Stuctures (in review)
 */

/**************** end self doc *******************************************/


int
main(int argc, char **argv)
{
	int i,j,ninf,npmax,npt;
	float r;
	float *xint,*zint;
	
	/* hook up getpar */
	initargs(argc, argv);
	requestdoc(0);
	
	/* get parameters */
	if (!getparint("ninf",&ninf))	ninf=5;
	if (!getparint("npmax",&npmax)) npmax=101;
	if (!getparfloat("r",&r)) r=0;
	
        checkpars();

	xint =  alloc1float(npmax);
	zint =  alloc1float(npmax);
	
	/* Input x z pairs specifying interfaces and velocity values. */
	for(i=0; i<ninf; ++i) {
		j= -1;
		do {
			j +=1;
			if(j>=npmax) 
		err("The point number on the %ith interface exceeds %i!",
				i+1,npmax);
			scanf("%f%f\n", &xint[j], &zint[j]);
		} while(zint[j]>-9999);
			npt = j;

		smooth_1(xint,zint,r,npt);
		for(j=0; j<npt; ++j)
 			printf("\t%g\t%g\n",xint[j],zint[j]);
 		printf("\t%g\t%g\n",1.,-99999.);
	}

	return(CWP_Exit());
}	       
